from abc import ABC, abstractmethod

import torch
from typing_extensions import override

from anytraverse import typing as anyt


class BaseHistory[TKey](ABC):
    """
    The base class to manage the history.
    """

    def __init__(self) -> None:
        self._store: list[anyt.HistoryElement[TKey]] = []

    @property
    def store(self) -> list[anyt.HistoryElement[TKey]]:
        return self._store

    @abstractmethod
    def find_best_match(
        self, query: TKey, *args, **kwargs
    ) -> anyt.HistoryElement[TKey]:
        pass

    def add(
        self, key: TKey, traversabilty_preferences: anyt.TraversabilityPreferences
    ) -> anyt.HistoryElement[TKey]:
        element: anyt.HistoryElement = (key, traversabilty_preferences)
        self._store.append(element)
        return element


class SimilarityHistory[TKey, TSim](BaseHistory[TKey], ABC):
    """
    A history that finds the best match based on a similarity function.
    """

    def __init__(self, similarity_func: anyt.SimilarityFunction[TKey, TSim]) -> None:
        super().__init__()
        self._similarity_func = similarity_func


class EncodingHistory(SimilarityHistory[anyt.Encoding, torch.Tensor]):
    """
    Stores the history as a pairs of encodings and traversability preferences.
    """

    def __init__(
        self,
        similarity_func: anyt.SimilarityFunction[anyt.Encoding, torch.Tensor],
    ) -> None:
        super().__init__(similarity_func=similarity_func)

    def _get_encodings(self) -> torch.Tensor:
        return torch.cat([encoding.view(1, -1) for (encoding, _) in self._store], dim=0)

    @override
    def find_best_match(
        self, query: anyt.Encoding
    ) -> anyt.HistoryElement[anyt.Encoding]:
        similarities = self._similarity_func(query, self._get_encodings()).ravel()
        best_match_inx = int(similarities.argmax())
        best_match = self._store[best_match_inx]
        return best_match



from abc import ABC, abstractmethod
from typing import Any
from anytraverse import typing as anyt


class BaseEncoder[TInput](ABC):
    """
    The base encoder module.

    Encodes any input to a numeric representation, as `torch.Tensor`.
    """

    DIM: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    @abstractmethod
    def __call__(
        self, x: TInput | list[TInput], *args: Any, **kwargs: Any
    ) -> anyt.Encoding:
        pass


class ImageEncoder[TImage: anyt.Image](BaseEncoder[TImage], ABC):
    """
    Base class for an image encoder module.
    """



import torch
from torch.nn import functional as F
from PIL import Image as PILImage
from transformers import AutoTokenizer, AutoModel, AutoImageProcessor
from anytraverse.utils.base.encoding import ImageEncoder
from anytraverse import typing as anyt
from anytraverse.helpers.device import DEVICE


class NomicImageEncoder[TImage: anyt.Image](ImageEncoder[TImage]):
    """
    Nomic image encoding model
    """

    DIM = 768

    def __init__(self, device: torch.device = DEVICE) -> None:
        super().__init__()
        self._processor = AutoImageProcessor.from_pretrained(
            "nomic-ai/nomic-embed-vision-v1.5",
            use_fast=True,
            cache_dir="data/weights/nomic",
        )
        self._model = AutoModel.from_pretrained(
            "nomic-ai/nomic-embed-vision-v1.5",
            trust_remote_code=True,
            cache_dir="data/weights/nomic",
        ).to(device=device)
        self._device = device

    def __call__(self, x: TImage | list[TImage]) -> anyt.Encoding:
        inputs = self._processor(x, return_tensors="pt").to(device=self._device)
        img_emb = self._model(**inputs).last_hidden_state
        return F.normalize(img_emb[:, 0], p=2, dim=1)



from anytraverse import typing as anyt


def update_traversability_preferences(
    traversability_preferences: anyt.TraversabilityPreferences,
    updates: anyt.TraversabilityPreferences,
) -> anyt.TraversabilityPreferences:
    """
    Updates the traversability preferences using the given udpates.

    Args:
        prefs (TraversabilityPreferences): The traversability preferences to update.
        updates (TraversabilityPreferences): The updates.

    Returns:
        TraversabilityPreferences:
            The updated traversability preferences.
    """
    return {**traversability_preferences, **updates}


def get_prompts(
    traversability_preferences: anyt.TraversabilityPreferences,
) -> list[anyt.Prompt]:
    return list(traversability_preferences.keys())


def get_weights(
    traversability_preferences: anyt.TraversabilityPreferences,
) -> list[anyt.Weight]:
    return list(traversability_preferences.values())


def parse_trav_pref_syntax(syntax: str) -> anyt.TraversabilityPreferences:
    """
    Parses traversability preference sytax to obtain a
    `TraversabilityPreferences` type `dict`.

    The syntax is:
    `prompt1: weight1; prompt2: weight2; ... ;`

    Args:
        syntax (str): The syntax describing the traversability preferences.

        Returns:
            TraversabilityPreferences:
                The traversability preferences described in the syntax as a `dict[str, float]`
    """
    syntax = syntax.strip()
    pws = syntax.split(";")
    prefs: anyt.TraversabilityPreferences = dict()
    for pw in pws:
        prompt, weight = pw.split(":")
        weight = float(weight)
        prefs[prompt.strip()] = weight
    return prefs



import torch
from PIL import Image as PILImage
import numpy as np
from numpy import typing as npt
from typing import Callable

type Prompt = str
type Weight = float
type TraversabilityPreferences = dict[Prompt, Weight]
type Image = PILImage.Image | npt.NDArray[np.uint8]
type Encoding = torch.Tensor
type TraversabilityMap = torch.Tensor
type PromptAttentionMap = torch.Tensor
type UncertaintyMap = torch.Tensor
type HistoryElement[TKey] = tuple[TKey, TraversabilityPreferences]
type SimilarityFunction[TElement, TSim] = Callable[[TElement, TElement], TSim]



from dataclasses import dataclass
from enum import Enum
from anytraverse import typing as anyt


class TraversalState(Enum):
    UNKOWN_OBJ = 1
    UNKNOWN_SCENE = 2
    OK = 3


@dataclass
class Threshold:
    ref_scene_similarity: float
    roi_uncertainty: float


@dataclass
class AnyTraverseState:
    """
    Contains the state of the AnyTraverse pipeline on one timestep,
    operating on an image.
    """

    image_encoding: anyt.Encoding
    """The image encoding of the image passed"""

    attention_maps: list[anyt.PromptAttentionMap]
    """
    The attention maps on the image for each of the prompts,
    each a `anyt.AttentionMap` (alias for `torch.Tensor`)
    of shape `(H, W)`, where the image passed in was had dimensions
    `W x H`
    """

    traversability_map: anyt.TraversabilityMap
    """
    The **traversability map**, a `anyt.TraversabilityMap` (alias
    for `torch.Tensor`) of shape `(H, W)` where each pixel gives the
    traversability score of that pixel between `0` (most untraversable)
    to `1` (most traversable)
    """

    uncertainty_map: anyt.UncertaintyMap
    """
    The uncertainty map, a `anyt.UncertaintyMap` (alias for `torch.Tensor`)
    of shape `(H, W)` where each pixel gives the uncertainty score of that
    pixel between `0` (most certain) to `1` (most uncertain)
    """

    traversability_map_roi: anyt.TraversabilityMap
    """
    The region of interest (ROI) extracted from the traversability map
    """

    uncertainty_map_roi: anyt.UncertaintyMap
    """
    The region of interest (ROI) extracted from the uncertainty map
    """

    ref_scene_similarity: float
    """
    Similarity score of current scene with the reference scene.
    """

    traversability_preferences: anyt.TraversabilityPreferences
    """
    The traversability preferences as `anyt.TraversabilityPreferences`
    (alias for `dict[str, float]`) of the form `{"<prompt1>": <weight1>, ...}`
    """

    roi_bbox: tuple[tuple[int, int], tuple[int, int]]
    """
    The region of interest (ROI), as start and end coordinates of a bounding box.
    Format: ((x_start, y_start), (x_end, y_end))
    """

    roi_uncertainty: float
    """The mean uncertainty in the region of interest in the image"""

    roi_traversability: float
    """The mean traversability in the region of interest in the image"""

    traversal_state: TraversalState
    """
    The state as a `TraversalState` `enum`, which contains whether it is
    ok (`OK`) to keep going, an unknown scene is encountered (`UNKNOWN_SCENE`),
    or an unknown object has entered the region of interest (`UNKNOWN_OBJ`)
    """
